part of secure_keyboard;

const double keyInputMonitorHeight = 50.0;
const double keyboardDefaultHeight = 280.0;
const int backspaceEventDelay = 100;

class SecureKeyboard extends StatefulWidget {
  final SecureKeyboardType type;
  final ValueChanged<SecureKeyboardKey> onKeyPressed;
  final ValueChanged<List<int>> onCharCodesChanged;
  final ValueChanged<List<int>> onConfirmKeyPressed;
  final VoidCallback onCloseKeyPressed;

  final String initText;
  final String hintText;
  final String inputTextLengthSymbol;
  final String confirmKeyText;
  final String clearKeyText;
  final String obscuringCharacter;
  final bool alwaysCaps;
  final bool obscureText;
  
  final double height;
  final Color backgroundColor;
  final Color stringKeyColor;
  final Color actionKeyColor;
  final Color confirmKeyColor;
  final TextStyle keyTextStyle;
  final TextStyle inputTextStyle;

  SecureKeyboard({
    Key key,
    @required this.type,
    @required this.onKeyPressed,
    @required this.onCharCodesChanged,
    @required this.onConfirmKeyPressed,
    @required this.onCloseKeyPressed,
    this.initText = '',
    this.hintText = '',
    this.inputTextLengthSymbol,
    this.confirmKeyText,
    this.clearKeyText,
    this.obscuringCharacter = '•',
    this.alwaysCaps = false,
    this.obscureText = true,
    this.height = keyboardDefaultHeight,
    this.backgroundColor = const Color(0xFF0A0A0A),
    this.stringKeyColor = const Color(0xFF313131),
    this.actionKeyColor = const Color(0xFF222222),
    this.confirmKeyColor = const Color(0xFF1C7CDC),
    this.keyTextStyle = const TextStyle(color: Colors.white, fontSize: 16.0, fontWeight: FontWeight.bold),
    this.inputTextStyle = const TextStyle(color: Colors.white, fontSize: 18.0, fontWeight: FontWeight.bold)
  })  : super(key: key);

  @override
  _SecureKeyboardState createState() => _SecureKeyboardState();
}

class _SecureKeyboardState extends State<SecureKeyboard> {
  final _methodChannel = const MethodChannel('flutter_secure_keyboard');

  final _definedKeyRows = List<List<SecureKeyboardKey>>();
  final _specialKeyRows = List<List<SecureKeyboardKey>>();
  final _charCodes = List<int>();
  
  Timer _backspaceEventGenerator;

  bool _isViewEnabled = false;
  bool _isShiftEnabled = false;
  bool _isSpecialCharsEnabled = false;
  
  void _initVariables() {
    _isViewEnabled = false;
    _isShiftEnabled = false;
    _isSpecialCharsEnabled = false;

    _definedKeyRows.clear();
    _specialKeyRows.clear();
    _charCodes.clear();
    _charCodes.addAll(widget.initText.codeUnits);

    if (widget.type == SecureKeyboardType.Numeric)
      _definedKeyRows.addAll(_getNumericKeyRows());
    else
      _definedKeyRows.addAll(_getAlphanumericKeyRows());

    _specialKeyRows.addAll(_getSpecialCharsKeyRows());
  }

  void _onKeyPressed(SecureKeyboardKey key) {
    if (key.type == SecureKeyboardKeyType.String) {
      final keyText = (_isShiftEnabled || widget.alwaysCaps)
          ? key.capsText
          : key.text;
      setState(() => _charCodes.add(keyText.codeUnits.first));
      widget.onCharCodesChanged(_charCodes);
    } else if (key.type == SecureKeyboardKeyType.Action) {
      switch (key.action) {
        // Backspace
        case SecureKeyboardKeyAction.Backspace:
          if (_charCodes.isNotEmpty) {
            setState(() => _charCodes.removeLast());
            widget.onCharCodesChanged(_charCodes);
          }
          break;
          
        // Confirm
        case SecureKeyboardKeyAction.Confirm:
          widget.onConfirmKeyPressed(_charCodes);
          break;
          
        // Clear
        case SecureKeyboardKeyAction.Clear:
          setState(() => _charCodes.clear());
          widget.onCharCodesChanged(_charCodes);
          break;
          
        // Shift
        case SecureKeyboardKeyAction.Shift:
          if (!widget.alwaysCaps)
            setState(() {
              _isShiftEnabled = !_isShiftEnabled;
            });
          break;
          
        // SpecialChars
        case SecureKeyboardKeyAction.SpecialChars:
          setState(() {
            _isSpecialCharsEnabled = !_isSpecialCharsEnabled;
          });
          break;

        default:
          return;
      }
    }

    widget.onKeyPressed(key);
  }

  @override
  void didUpdateWidget(covariant SecureKeyboard oldWidget) {
    super.didUpdateWidget(oldWidget);
    _initVariables();
  }

  @override
  void initState() {
    super.initState();
    _methodChannel.invokeMethod('secureModeOn');
    _initVariables();
  }

  @override
  Widget build(BuildContext context) {
    final keyRows = _isSpecialCharsEnabled
        ? _specialKeyRows
        : _definedKeyRows;
    final keyboardKey = _buildKeyboardKey(keyRows);
    keyboardKey.insert(0, _buildKeyInputMonitor());

    return Container(
      width: MediaQuery.of(context).size.width,
      height: widget.height + keyInputMonitorHeight,
      color: widget.backgroundColor,
      child: Column(
        mainAxisAlignment: MainAxisAlignment.spaceBetween,
        crossAxisAlignment: CrossAxisAlignment.center,
        children: keyboardKey
      ),
    );
  }

  @override
  void dispose() {
    _methodChannel.invokeMethod('secureModeOff');
    super.dispose();
  }

  Widget _buildKeyInputMonitor() {
    String secureText;
    TextStyle secureTextStyle;

    if (_charCodes.isNotEmpty) {
      if (widget.obscureText && !_isViewEnabled) {
        secureText = '';
        for (var i=0; i<_charCodes.length; i++) {
          if (i == _charCodes.length - 1)
            secureText += String.fromCharCode(_charCodes[i]);
          else
            secureText += widget.obscuringCharacter;
        }
      } else {
        secureText = String.fromCharCodes(_charCodes);
      }

      secureTextStyle = widget.inputTextStyle;
    } else {
      secureText = widget.hintText;
      secureTextStyle = widget.inputTextStyle.copyWith(
          color: widget.inputTextStyle.color.withOpacity(0.5));
    }

    final lengthSymbol = widget.inputTextLengthSymbol ?? (Platform.localeName == 'ko_KR')
        ? '자'
        : 'digit';
    final lengthText = '${_charCodes.length}$lengthSymbol';

    Widget viewKey = SizedBox();
    if (widget.obscureText) {
      viewKey = Container(
        width: keyInputMonitorHeight / 1.5,
        height: keyInputMonitorHeight / 1.5,
        margin: const EdgeInsets.only(left: 1.5),
        child: Material(
          color: Colors.transparent,
          child: InkWell(
            borderRadius: BorderRadius.circular(24.0),
            child: GestureDetector(
              onTapDown: (details) => setState(() => _isViewEnabled = true),
              onTapUp: (details) => setState(() => _isViewEnabled = false),
              child: Icon(Icons.remove_red_eye, color: widget.keyTextStyle.color)
            )
          ),
        ),
      );
    }

    return SizedBox(
      height: keyInputMonitorHeight,
      child: Row(
        children: [
          Expanded(
            child: Padding(
              padding: const EdgeInsets.only(left: 8.0),
              child: Text(secureText, style: secureTextStyle)
            )
          ),
          Padding(
            padding: const EdgeInsets.symmetric(horizontal: 5.0),
            child: Text(lengthText, style: widget.keyTextStyle)
          ),
          viewKey,
          Container(
            width: keyInputMonitorHeight / 1.5,
            height: keyInputMonitorHeight / 1.5,
            margin: const EdgeInsets.only(right: 1.5),
            child: Material(
              color: Colors.transparent,
              child: InkWell(
                borderRadius: BorderRadius.circular(24.0),
                onTap: widget.onCloseKeyPressed,
                child: Icon(Icons.close, color: widget.keyTextStyle.color)
              ),
            ),
          )
        ],
      ),
    );
  }

  List<Widget> _buildKeyboardKey(List<List<SecureKeyboardKey>> keyRows) {
    return List.generate(keyRows.length, (int rowNum) {
      return Material(
        color: Colors.transparent,
        child: Row(
          mainAxisAlignment: MainAxisAlignment.spaceAround,
          crossAxisAlignment: CrossAxisAlignment.center,
          children: List.generate(keyRows[rowNum].length, (int keyNum) {
            final key = keyRows[rowNum][keyNum];
            
            switch (key.type) {
              case SecureKeyboardKeyType.String:
                return _buildStringKey(key, keyRows.length);
              case SecureKeyboardKeyType.Action:
                return _buildActionKey(key, keyRows.length);
              default:
                throw Exception('Unknown key type.');
            }
          })
        ),
      );
    });
  }
  
  Widget _buildStringKey(SecureKeyboardKey key, int keyRowsLength) {
    final keyText = (_isShiftEnabled || widget.alwaysCaps)
        ? key.capsText
        : key.text;

    return Expanded(
      child: Container(
        height: widget.height / keyRowsLength,
        padding: const EdgeInsets.all(1.5),
        child: Material(
          borderRadius: BorderRadius.circular(4.0),
          color: widget.stringKeyColor,
          child: InkWell(
            onTap: () => _onKeyPressed(key),
            child: Center(child: Text(keyText, style: widget.keyTextStyle))
          ),
        ),
      ),
    );
  }

  Widget _buildActionKey(SecureKeyboardKey key, int keyRowsLength) {
    Widget actionKey;

    switch (key.action) {
      case SecureKeyboardKeyAction.Backspace:
        actionKey = GestureDetector(
          onLongPress: () {
            final delay = Duration(milliseconds: backspaceEventDelay);
            _backspaceEventGenerator = Timer.periodic(delay, (_) => _onKeyPressed(key));
          },
          onLongPressUp: () {
            if (_backspaceEventGenerator != null) {
              _backspaceEventGenerator.cancel();
              _backspaceEventGenerator = null;
            }
          },
          child: Icon(Icons.backspace, color: widget.keyTextStyle.color)
        );
        break;
      case SecureKeyboardKeyAction.Shift:
        actionKey = Icon(Icons.arrow_upward, color: widget.keyTextStyle.color);
        break;
      case SecureKeyboardKeyAction.Clear:
        actionKey = Text(
          widget.clearKeyText ?? (Platform.localeName == 'ko_KR')
              ? '초기화'
              : 'Clear',
          style: widget.keyTextStyle
        );
        break;
      case SecureKeyboardKeyAction.Confirm:
        actionKey = Text(
          widget.confirmKeyText ?? (Platform.localeName == 'ko_KR')
              ? '입력완료'
              : 'Confirm',
          style: widget.keyTextStyle
        );
        break;
      case SecureKeyboardKeyAction.SpecialChars:
        actionKey = Text(
          _isSpecialCharsEnabled
              ? (_isShiftEnabled ? 'ABC' : 'abc')
              : '!@#',
          style: widget.keyTextStyle
        );
        break;
      case SecureKeyboardKeyAction.Blank:
        return Expanded(child: SizedBox());
    }

    return Expanded(
      child: Container(
        height: widget.height / keyRowsLength,
        padding: const EdgeInsets.all(1.5),
        child: Material(
          borderRadius: BorderRadius.circular(4.0),
          color: (key.action == SecureKeyboardKeyAction.Confirm)
              ? widget.confirmKeyColor
              : widget.actionKeyColor,
          child: InkWell(
            onTap: () => _onKeyPressed(key),
            child: Center(child: actionKey)
          ),
        ),
      ),
    );
  }
}